from dateutil.relativedelta import relativedelta
from edc_visit_schedule.constants import DAY1, MONTH1, MONTH3, MONTH6, WEEK2
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.tests import DummyPanel
from edc_visit_schedule.visit import Crf, CrfCollection, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule


def get_visit_schedule(cdef):

    app_label = "edc_metadata"

    class MockPanel(DummyPanel):
        """`requisition_model` is normally set when the lab profile
        is set up.
        """

        def __init__(self, name):
            super().__init__(requisition_model="edc_metadata.subjectrequisition", name=name)

    crfs_prn = CrfCollection(
        Crf(show_order=100, model=f"{app_label}.prnone"),
        Crf(show_order=200, model=f"{app_label}.prntwo"),
    )

    crfs_missed = CrfCollection(
        Crf(show_order=1, model="edc_metadata.subjectvisitmissed", required=True),
    )

    crfs0 = CrfCollection(Crf(show_order=1, model=f"{app_label}.crftwo", required=False))
    crfs1 = CrfCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=True))
    crfs2 = CrfCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=False))
    crfs3 = CrfCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=False))
    crfs4 = CrfCollection(Crf(show_order=1, model=f"{app_label}.crfone", required=False))

    visit0 = Visit(
        code=DAY1,
        title="Week 1",
        timepoint=0.0,
        rbase=relativedelta(days=0),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs0,
        crfs_prn=crfs_prn,
        allow_unscheduled=True,
        facility_name="5-day-clinic",
    )

    visit1 = Visit(
        code=WEEK2,
        title="Week 2",
        timepoint=1.0,
        rbase=relativedelta(days=7),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs1,
        crfs_prn=crfs_prn,
        crfs_missed=crfs_missed,
        facility_name="5-day-clinic",
    )

    visit2 = Visit(
        code=MONTH1,
        title="Month 1",
        timepoint=2.0,
        rbase=relativedelta(months=1),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs2,
        crfs_prn=crfs_prn,
        facility_name="5-day-clinic",
    )

    visit3 = Visit(
        code=MONTH3,
        title="Month 3",
        timepoint=3.0,
        rbase=relativedelta(months=3),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs3,
        crfs_prn=crfs_prn,
        facility_name="5-day-clinic",
    )

    visit4 = Visit(
        code=MONTH6,
        title="Month 6",
        timepoint=4.0,
        rbase=relativedelta(months=6),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        crfs=crfs4,
        crfs_prn=crfs_prn,
        facility_name="5-day-clinic",
    )

    schedule = Schedule(
        name="schedule",
        onschedule_model="edc_metadata.onschedule",
        offschedule_model="edc_metadata.offschedule",
        consent_definitions=[cdef],
        appointment_model="edc_appointment.appointment",
    )

    schedule.add_visit(visit0)
    schedule.add_visit(visit1)
    schedule.add_visit(visit2)
    schedule.add_visit(visit3)
    schedule.add_visit(visit4)

    visit_schedule = VisitSchedule(
        name="visit_schedule",
        offstudy_model="edc_offstudy.subjectoffstudy",
        death_report_model="edc_metadata.deathreport",
    )

    visit_schedule.add_schedule(schedule)
    return visit_schedule



from dateutil.relativedelta import relativedelta
from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.tests import DummyPanel
from edc_visit_schedule.visit import (
    Crf,
    CrfCollection,
    Requisition,
    RequisitionCollection,
    Visit,
)
from edc_visit_schedule.visit_schedule import VisitSchedule

app_label = "edc_metadata"


def get_visit_schedule(cdef):

    class MockPanel(DummyPanel):
        """`requisition_model` is normally set when the lab profile
        is set up.
        """

        def __init__(self, name):
            super().__init__(requisition_model="edc_metadata.subjectrequisition", name=name)

    crfs_prn = CrfCollection(
        Crf(show_order=100, model=f"{app_label}.prnone"),
        Crf(show_order=200, model=f"{app_label}.prntwo"),
    )

    crfs_missed = CrfCollection(
        Crf(show_order=1000, model="edc_metadata.subjectvisitmissed", required=True),
    )

    crfs0 = CrfCollection(
        Crf(show_order=1, model=f"{app_label}.crfone", required=True),
        Crf(show_order=2, model=f"{app_label}.crftwo", required=True),
        Crf(show_order=3, model=f"{app_label}.crfthree", required=True),
        Crf(show_order=4, model=f"{app_label}.crffour", required=True),
        Crf(show_order=5, model=f"{app_label}.crffive", required=True),
    )

    crfs1 = CrfCollection(
        Crf(show_order=1, model=f"{app_label}.crffour", required=True),
        Crf(show_order=2, model=f"{app_label}.crffive", required=True),
        Crf(show_order=3, model=f"{app_label}.crfsix", required=True),
    )

    crfs2 = CrfCollection(Crf(show_order=1, model=f"{app_label}.crfseven", required=True))

    crfs_unscheduled = CrfCollection(
        Crf(show_order=1, model=f"{app_label}.crftwo", required=True),
        Crf(show_order=2, model=f"{app_label}.crfthree", required=True),
        Crf(show_order=3, model=f"{app_label}.crffive", required=True),
    )

    requisitions = RequisitionCollection(
        Requisition(show_order=10, panel=MockPanel("one"), required=True, additional=False),
        Requisition(show_order=20, panel=MockPanel("two"), required=True, additional=False),
        Requisition(show_order=30, panel=MockPanel("three"), required=False, additional=False),
        Requisition(show_order=40, panel=MockPanel("four"), required=False, additional=False),
        Requisition(show_order=50, panel=MockPanel("five"), required=False, additional=False),
        Requisition(show_order=60, panel=MockPanel("six"), required=False, additional=False),
        Requisition(show_order=70, panel=MockPanel("seven"), required=False, additional=False),
        Requisition(show_order=80, panel=MockPanel("eight"), required=False, additional=False),
    )

    requisitions3000 = RequisitionCollection(
        Requisition(show_order=10, panel=MockPanel("seven"), required=True, additional=False)
    )

    requisitions_unscheduled = RequisitionCollection(
        Requisition(show_order=10, panel=MockPanel("one"), required=True, additional=False),
        Requisition(show_order=20, panel=MockPanel("three"), required=True, additional=False),
        Requisition(show_order=30, panel=MockPanel("five"), required=True, additional=False),
        Requisition(show_order=90, panel=MockPanel("nine"), required=True, additional=False),
    )

    visit0 = Visit(
        code="1000",
        title="Week 1",
        timepoint=0,
        rbase=relativedelta(days=0),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        requisitions=requisitions,
        crfs=crfs0,
        crfs_unscheduled=crfs_unscheduled,
        crfs_prn=crfs_prn,
        requisitions_unscheduled=requisitions_unscheduled,
        allow_unscheduled=True,
        facility_name="5-day-clinic",
    )

    visit1 = Visit(
        code="2000",
        title="Week 2",
        timepoint=1,
        rbase=relativedelta(days=7),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        requisitions=requisitions,
        crfs=crfs1,
        crfs_prn=crfs_prn,
        crfs_missed=crfs_missed,
        facility_name="5-day-clinic",
    )

    visit2 = Visit(
        code="3000",
        title="Week 3",
        timepoint=2,
        rbase=relativedelta(days=14),
        rlower=relativedelta(days=0),
        rupper=relativedelta(days=6),
        requisitions=requisitions3000,
        crfs=crfs2,
        crfs_prn=crfs_prn,
        facility_name="5-day-clinic",
    )

    schedule = Schedule(
        name="schedule",
        onschedule_model="edc_metadata.onschedule",
        offschedule_model="edc_metadata.offschedule",
        consent_definitions=[cdef],
        appointment_model="edc_appointment.appointment",
    )

    schedule.add_visit(visit0)
    schedule.add_visit(visit1)
    schedule.add_visit(visit2)

    visit_schedule = VisitSchedule(
        name="visit_schedule",
        offstudy_model="edc_offstudy.subjectoffstudy",
        death_report_model="edc_metadata.deathreport",
    )

    visit_schedule.add_schedule(schedule)
    return visit_schedule



from datetime import datetime
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_consent import site_consents
from edc_consent.consent_definition import ConsentDefinition
from edc_constants.constants import FEMALE, MALE
from edc_facility import import_holidays
from edc_lab.models import Panel
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ...models import CrfMetadata, RequisitionMetadata
from ..models import SubjectConsentV1
from ..visit_schedule import get_visit_schedule

test_datetime = datetime(2019, 6, 11, 8, 00, tzinfo=ZoneInfo("UTC"))


class TestMetadataMixin(TestCase):
    @classmethod
    def setUpTestData(cls):
        import_holidays()

    def setUp(self):
        traveller = time_machine.travel(test_datetime)
        traveller.start()
        self.panel_one = Panel.objects.create(name="one")
        self.panel_two = Panel.objects.create(name="two")

        for name in ["three", "four", "five", "six"]:
            Panel.objects.create(name=name)

        consent_v1 = ConsentDefinition(
            "edc_metadata.subjectconsentv1",
            version="1",
            start=get_utcnow(),
            end=get_utcnow() + relativedelta(years=3),
            age_min=18,
            age_is_adult=18,
            age_max=64,
            gender=[MALE, FEMALE],
        )

        site_consents.registry = {}
        site_consents.register(consent_v1)

        site_visit_schedules._registry = {}
        site_visit_schedules.loaded = False
        site_visit_schedules.register(get_visit_schedule(consent_v1))
        self.subject_identifier = "1111111"
        self.assertEqual(CrfMetadata.objects.all().count(), 0)
        self.assertEqual(RequisitionMetadata.objects.all().count(), 0)
        subject_consent = SubjectConsentV1.objects.create(
            subject_identifier=self.subject_identifier, consent_datetime=get_utcnow()
        )
        _, self.schedule = site_visit_schedules.get_by_onschedule_model(
            "edc_metadata.onschedule"
        )
        self.schedule.put_on_schedule(
            subject_identifier=self.subject_identifier,
            onschedule_datetime=subject_consent.consent_datetime,
        )
        self.appointment = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code=self.schedule.visits.first.code,
        )

        self.appointment_2000 = Appointment.objects.get(
            subject_identifier=self.subject_identifier,
            visit_code="2000",
        )
        traveller.stop()



from django.apps import apps as django_apps
from edc_auth.site_auths import site_auths

if django_apps.is_installed("edc_export"):
    from edc_export.constants import EXPORT

    site_auths.update_group(
        "edc_metadata.export_crfmetadata",
        "edc_metadata.export_requisitionmetadata",
        name=EXPORT,
    )



from datetime import datetime
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, override_settings
from edc_visit_tracking.constants import SCHEDULED

from edc_metadata.constants import KEYED, NOT_REQUIRED, REQUIRED
from edc_metadata.metadata_handler import MetadataHandlerError
from edc_metadata.metadata_inspector import MetaDataInspector
from edc_metadata.metadata_updater import MetadataUpdater
from edc_metadata.models import CrfMetadata, RequisitionMetadata

from ..models import CrfOne, CrfThree, CrfTwo, SubjectRequisition, SubjectVisit
from .metadata_test_mixin import TestMetadataMixin

test_datetime = datetime(2019, 6, 11, 8, 00, tzinfo=ZoneInfo("UTC"))


@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=test_datetime - relativedelta(years=3),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=test_datetime + relativedelta(years=3),
)
class TestMetadataUpdater(TestMetadataMixin, TestCase):
    def test_updates_crf_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment,
            subject_identifier=self.subject_identifier,
            report_datetime=self.appointment.appt_datetime,
            visit_code=self.appointment.visit_code,
            visit_code_sequence=self.appointment.visit_code_sequence,
            visit_schedule_name=self.appointment.visit_schedule_name,
            schedule_name=self.appointment.schedule_name,
            reason=SCHEDULED,
        )
        CrfOne.objects.create(subject_visit=subject_visit)
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=KEYED,
                model="edc_metadata.crfone",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crftwo",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crfthree",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_updates_all_crf_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        CrfOne.objects.create(subject_visit=subject_visit)
        CrfTwo.objects.create(subject_visit=subject_visit)
        CrfThree.objects.create(subject_visit=subject_visit)
        for model_name in ["crfone", "crftwo", "crfthree"]:
            self.assertEqual(
                CrfMetadata.objects.filter(
                    entry_status=KEYED,
                    model=f"edc_metadata.{model_name}",
                    visit_code=subject_visit.visit_code,
                ).count(),
                1,
            )

    def test_updates_requisition_metadata_as_keyed(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        SubjectRequisition.objects.create(subject_visit=subject_visit, panel=self.panel_one)
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=KEYED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_one.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_two.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_resets_crf_metadata_on_delete(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        crf_one = CrfOne.objects.create(subject_visit=subject_visit)
        crf_one.delete()
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crfone",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crftwo",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            CrfMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.crfthree",
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_resets_requisition_metadata_on_delete1(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        obj = SubjectRequisition.objects.create(
            subject_visit=subject_visit, panel=self.panel_one
        )
        obj.delete()
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_one.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_two.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_resets_requisition_metadata_on_delete2(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        obj = SubjectRequisition.objects.create(
            subject_visit=subject_visit, panel=self.panel_two
        )
        obj.delete()
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_one.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )
        self.assertEqual(
            RequisitionMetadata.objects.filter(
                entry_status=REQUIRED,
                model="edc_metadata.subjectrequisition",
                panel_name=self.panel_two.name,
                visit_code=subject_visit.visit_code,
            ).count(),
            1,
        )

    def test_get_metadata_for_subject_visit(self):
        """Asserts can get metadata for a subject and visit code."""
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        self.assertEqual(len(subject_visit.visit.all_crfs), 7)
        self.assertEqual(len(subject_visit.visit.all_requisitions), 9)

        metadata_a = []
        for key, values in subject_visit.metadata.items():
            for obj in values:
                try:
                    metadata_a.append(f"{obj.model}.{obj.panel_name}")
                except AttributeError:
                    metadata_a.append(obj.model)
        metadata_a.sort()
        forms = (
            subject_visit.schedule.visits.get(subject_visit.visit_code).scheduled_forms.forms
            + subject_visit.schedule.visits.get(subject_visit.visit_code).prn_forms.forms
        )
        metadata_b = [f.full_name for f in forms]
        metadata_b = list(set(metadata_b))
        metadata_b.sort()
        self.assertEqual(metadata_a, metadata_b)

    def test_metadata_inspector(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        inspector = MetaDataInspector(
            model_cls=CrfOne,
            visit_schedule_name=subject_visit.visit_schedule_name,
            schedule_name=subject_visit.schedule_name,
            visit_code=subject_visit.visit_code,
            timepoint=subject_visit.timepoint,
        )
        self.assertEqual(len(inspector.required), 1)
        self.assertEqual(len(inspector.keyed), 0)

        CrfOne.objects.create(subject_visit=subject_visit)

        inspector = MetaDataInspector(
            model_cls=CrfOne,
            visit_schedule_name=subject_visit.visit_schedule_name,
            schedule_name=subject_visit.schedule_name,
            visit_code=subject_visit.visit_code,
            timepoint=subject_visit.timepoint,
        )
        self.assertEqual(len(inspector.required), 0)
        self.assertEqual(len(inspector.keyed), 1)

    def test_crf_updates_ok(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        CrfMetadata.objects.get(
            visit_code=subject_visit.visit_code,
            model="edc_metadata.crfone",
            entry_status=REQUIRED,
        )
        metadata_updater = MetadataUpdater(
            related_visit=subject_visit,
            source_model="edc_metadata.crfone",
        )
        metadata_updater.get_and_update(entry_status=NOT_REQUIRED)
        self.assertRaises(
            ObjectDoesNotExist,
            CrfMetadata.objects.get,
            visit_code=subject_visit.visit_code,
            model="edc_metadata.crfone",
            entry_status=REQUIRED,
        )

        for visit_obj in SubjectVisit.objects.all():
            if visit_obj == subject_visit:
                try:
                    CrfMetadata.objects.get(
                        visit_code=visit_obj.visit_code,
                        model="edc_metadata.crfone",
                        entry_status=NOT_REQUIRED,
                    )
                except ObjectDoesNotExist as e:
                    self.fail(e)
            else:
                self.assertRaises(
                    ObjectDoesNotExist,
                    CrfMetadata.objects.get,
                    visit_code=visit_obj.visit_code,
                    model="edc_metadata.crfone",
                    entry_status=NOT_REQUIRED,
                )

    def test_crf_invalid_model(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        metadata_updater = MetadataUpdater(
            related_visit=subject_visit,
            source_model="edc_metadata.blah",
        )
        self.assertRaises(
            MetadataHandlerError, metadata_updater.get_and_update, entry_status=NOT_REQUIRED
        )

    def test_crf_model_not_scheduled(self):
        subject_visit = SubjectVisit.objects.create(
            appointment=self.appointment, reason=SCHEDULED
        )
        metadata_updater = MetadataUpdater(
            related_visit=subject_visit,
            source_model="edc_metadata.crfseven",
        )
        self.assertRaises(
            MetadataHandlerError,
            metadata_updater.get_and_update,
            entry_status=NOT_REQUIRED,
        )



from datetime import datetime
from zoneinfo import ZoneInfo

from dateutil.relativedelta import relativedelta
from edc_consent.consent_definition import ConsentDefinition
from edc_constants.constants import FEMALE, MALE

test_datetime = datetime(2019, 6, 11, 8, 00, tzinfo=ZoneInfo("UTC"))

consent_v1 = ConsentDefinition(
    "edc_metadata.subjectconsentv1",
    version="1",
    start=test_datetime,
    end=test_datetime + relativedelta(years=3),
    age_min=18,
    age_is_adult=18,
    age_max=64,
    gender=[MALE, FEMALE],
)
